###IMPORT MODULES

from __future__ import print_function, division
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

num_epochs = 100
total_series_length = 50000
truncated_backprop_length = 15
state_size = 4
num_classes = 2
echo_step = 3
batch_size = 5
num_batches = total_series_length//batch_size//truncated_backprop_length


###GENERATE DATA AND LABELS
def generateData():
    x = np.array(np.random.choice(2, total_series_length, p=[0.5, 0.5]))
    y = np.roll(x, echo_step)
    y[0:echo_step] = 0
    
    #reshape the generated data so that you can train in batches
    x = x.reshape((batch_size, -1))  # The first index changing slowest, subseries as rows
    y = y.reshape((batch_size, -1)) #note that in python, the -1 for the second entry in the reshape command means it automatically calculates the second entry for the array shape based on the first entry
    
    return (x, y)


###INITIALIZE PLACEHOLDERS, I.E., THINGS THAT WILL CHANGE DURING TRAINING WITH THE BATCH
#create placeholders for the input and output for each batch
#batch_size is the number of simultaneous sequences being processed
#truncated_backprop_length is the size of the chunks from the data batch that your are backpropping through
batchX_placeholder = tf.placeholder(tf.float32, [batch_size, truncated_backprop_length])
batchY_placeholder = tf.placeholder(tf.int32, [batch_size, truncated_backprop_length])

#the first dimension of initial state is of size batch_size, since you need batch_size instances of states when propagating forward
init_state = tf.placeholder(tf.float32, [batch_size, state_size])
#INITIALIZE THE WEIGHTS AND BIASES
#weights and bias for the output
W2 = tf.Variable(np.random.rand(state_size, num_classes),dtype=tf.float32)
b2 = tf.Variable(np.zeros((1,num_classes)), dtype=tf.float32)

# Unpack columns, note that I changed the tf.unpack to tf.unstack here, since tf.unpack is deprecated in TF 1.3
inputs_series = tf.split(batchX_placeholder, truncated_backprop_length,axis=1) #note that I changed the order of the inputs to tf.split to correspond to TF version 1.3
labels_series = tf.unstack(batchY_placeholder, axis=1)

###FORWARD PASS
cell = tf.nn.rnn_cell.BasicRNNCell(state_size)
states_series, current_state = tf.nn.static_rnn(cell=cell,inputs=inputs_series,initial_state=init_state)

###CALCUALTE LOSSES AND PREDICTIONS, DEFINE THE TRAINING STEP

#multiply the state by the output matrix and add the bias
#do this for every element of the output sequence
logits_series = [tf.matmul(state, W2) + b2 for state in states_series] #Broadcasted addition
#apply a softmax for each of the output vectors to normalize them to probability vectors
predictions_series = [tf.nn.softmax(logits) for logits in logits_series]

#apply the softmax cross entropy loss, note that we input the logits because the softmax function here applies a softmax, as the name implies
#note that sparse softmax is used when the true label is given as a 1-hot representation, e.g., a sparse label
losses = [tf.nn.sparse_softmax_cross_entropy_with_logits(labels=labels,logits=logits) for labels, logits in zip(labels_series,logits_series)] #note that I changed this since you need to write labels=labels and logits=logits, not just labels,logits inside the sparse_softmax_.... function
#also note that the usage is to have logits is of shape [batch_size, num_classes] and labels of shape [batch_size]
#in other word, each entry in labels must be an index in [0, num_classes)
total_loss = tf.reduce_mean(losses)

#train with Adagrad to minimize the loss
train_step = tf.train.AdagradOptimizer(0.3).minimize(total_loss)

###DEFINE A FUNCTION FOR PLOTTING THE LOSS VS TIME, AND THE DATA,LABELS, AND PREDICTIONS FOR 5 SAMPLES IN A BATCH
def plot(loss_list, predictions_series, batchX, batchY):
    plt.subplot(2, 3, 1)
    plt.cla()
    #plot the loss over time
    plt.plot(loss_list)
    
    for batch_series_idx in range(5): #the 5 here is the batch_size, so make one plot for each sample in the batch
        one_hot_output_series = np.array(predictions_series)[:, batch_series_idx, :] #2x15 probabilities for one batch
        single_output_series = np.array([(1 if out[0] < 0.5 else 0) for out in one_hot_output_series])#2x15 one hot representation
        
        plt.subplot(2, 3, batch_series_idx + 2)
        plt.cla()
        plt.axis([0, truncated_backprop_length, 0, 2]) #set the x-axis going from 0 to backprop length
        left_offset = range(truncated_backprop_length) #make a x-axis vector going from 0 to backprop length for plotting
        plt.bar(left_offset, batchX[batch_series_idx, :], width=1, color="blue") #blue is the input, either on or off at height 1
        plt.bar(left_offset, batchY[batch_series_idx, :] * 0.5, width=1, color="red") #red is the true labels, either on or off at height 0.5
        plt.bar(left_offset, single_output_series * 0.3, width=1, color="green") #green are the predicted labels, either on or off at height 0.3
    
    plt.draw()
    plt.pause(0.0001)

###RUN THE SESSION
with tf.Session() as sess:
    sess.run(tf.initialize_all_variables())
    plt.ion()#turn on interactive plot mode, so that you can keep modifying the plot in time
    plt.figure() #open a figure
    plt.show() #show the plot
    loss_list = [] #make a list of losses to plot
    
    #perform a certain number of epochs
    for epoch_idx in range(num_epochs):
        x,y = generateData() #generate the data, note that this isn't the usual way to do it, but it still works for this example since the function we learn in each epoch is the same
        _current_state = np.zeros((batch_size, state_size)) #initialize the current hidden state to zero, for all the batches
        
        print("New data, epoch", epoch_idx)
        
        for batch_idx in range(num_batches):
            #define the start and end indexes going through chunks of shape "batch_size X truncated_backprop_length"
            start_idx = batch_idx * truncated_backprop_length
            end_idx = start_idx + truncated_backprop_length
            
            batchX = x[:,start_idx:end_idx]
            batchY = y[:,start_idx:end_idx]
            
            #note that when training, the _current_state is updated and fed back into the intial state of the next batch
            #so, at the end of the training its not surprising that (in the plot) all of the predictions are correct, including the first few entries
            _total_loss, _train_step, _current_state, _predictions_series = sess.run(
             [total_loss, train_step, current_state, predictions_series],
             feed_dict={
                 batchX_placeholder:batchX,
                 batchY_placeholder:batchY,
                 init_state:_current_state
             })
             
            loss_list.append(_total_loss)
             
            if batch_idx%100 == 0:
                print("Step",batch_idx, "Loss", _total_loss)
                plot(loss_list, _predictions_series, batchX, batchY)

plt.ioff()
plt.show()